AlmanahTagEntry {
   color: #aaa;
   font-size: 8px;
   transition: none;
}

AlmanahTagEntry:focus {
   color: @theme_fg_color;
}

.almanah-mw-scrolled {
    border-bottom-width: 1px;
    border-top-width: 0px;
    border-left-width: 0px;
    border-right-width: 0px;
}

.almanah-events-scrolled {
    border-bottom-width: 0px;
    border-top-width: 1px;
    border-left-width: 0px;
    border-right-width: 0px;
}
